import csv
import itertools
import re
from pathlib import Path
from typing import *

import pandas
from loguru import logger

NUMERIC_REGEX = re.compile("^.?(?P<number>[\d]+)")


def checkdir(path: Union[str, Path]) -> Path:
	path = Path(path)
	if not path.exists():
		path.mkdir()
	return path


def get_numeric_columns(columns: List[Union[str, int, float]]) -> List[Union[str, int, float]]:
	numeric_columns = list()
	for column in columns:
		if isinstance(column, str):
			match = NUMERIC_REGEX.search(column)
			if match:
				col = match.groupdict()['number']
			else:
				continue
		else:
			col = column

		try:
			int(col)
		except (ValueError, TypeError):
			continue
		numeric_columns.append(column)
	return numeric_columns


def map_trajectories_to_genotype(genotype_members: pandas.Series) -> Dict[str, str]:
	""" Maps each trajectory to the genotype it belongs to."""
	trajectory_to_genotype = dict()
	for genotype_label, members in genotype_members.items():
		for member in members.split('|'):
			trajectory_to_genotype[member] = genotype_label
	return trajectory_to_genotype


def get_valid_points(left_trajectory: pandas.Series, right_trajectory: pandas.Series, dlimit: float, flimit: Optional[float] = None,
		inner: bool = False) -> Tuple[pandas.Series, pandas.Series]:
	"""
		Filters out timepoints that do not satisfy the detection criteria.
	Parameters
	----------
	left_trajectory: pandas.Series
	right_trajectory: pandas.Series
	dlimit: float
		Removes the timepoint if both points do no exceed this value.
	flimit: float
		If given, all points exceeding this value are treated as "undetected".
	inner: bool; default False
		If `true`, both points must exceed the detection limit for a given timepoint to be considered valid.

	Returns
	-------

	"""

	if flimit is not None:
		# Mask the values so they are considered below the detection limit.
		# Assign a value of -1 so that they will be excluded even if the detection limit is 0.
		# list comprehension is ~6X faster than mask()
		left = pandas.Series([(-1 if s > 0.97 else s) for s in left_trajectory.values], index = left_trajectory.index)
		right = pandas.Series([(-1 if s > 0.97 else s) for s in right_trajectory.values], index = right_trajectory.index)
	else:
		left, right = left_trajectory, right_trajectory

	if inner:
		# list comprehensions are ~10X faster than using the built-in pandas methods.
		at_least_one_detected = [(l > dlimit and r > dlimit) for l, r in zip(left.values, right.values)]
	else:
		at_least_one_detected = [(l > dlimit or r > dlimit) for l, r in zip(left.values, right.values)]
	at_least_one_detected = pandas.Series(at_least_one_detected, index = left.index)

	# Remove indicies where the series value falls below the detection limit. This should include the masked fixed values.
	at_least_one_detected_reduced = at_least_one_detected[at_least_one_detected]
	if at_least_one_detected_reduced.empty:
		# There are no shared timepoints between the series. Assign index_min and index_max to the same number, which will result in an empty dataframe.
		position_index_min = position_index_max = 0
	else:
		# Apparently the min() and max functions now work with strings as well as numbers.
		# Cast the numbers to float so the typeerror is thrown correctly.
		try:
			position_index_min_value = min(at_least_one_detected_reduced.index, key = lambda s: float(s))
			position_index_max_value = max(at_least_one_detected_reduced.index, key = lambda s: float(s))
		except TypeError:
			# The indicies are str and we can't use min() or max(). Assume the indicies are already sorted.
			position_index_min_value = at_least_one_detected_reduced.index[0]
			position_index_max_value = at_least_one_detected_reduced.index[-1]

		position_index_min = left.index.get_loc(position_index_min_value)
		position_index_max = left.index.get_loc(position_index_max_value)
		# Since we want to include the last index, increment position_index_max by one.
		position_index_max += 1
	result_left = left_trajectory[position_index_min:position_index_max]
	result_right = right_trajectory[position_index_min:position_index_max]

	return result_left, result_right


get_detected_points = get_valid_points


def calculate_luminance(color: str) -> float:
	# 0.299 * color.R + 0.587 * color.G + 0.114 * color.B
	red = int(color[1:3], 16)
	green = int(color[3:5], 16)
	blue = int(color[5:], 16)

	lum = (.299 * red) + (.587 * green) + (.114 * blue)
	return lum / 255


def format_inconsistency_matrix(inconsistency_matrix) -> pandas.DataFrame:
	inconsistency_table = pandas.DataFrame(inconsistency_matrix, columns = ['mean', 'std', 'observations', 'statistic'])
	inconsistency_table['observations'] = inconsistency_table['observations'].astype(int)
	return inconsistency_table


# noinspection PyTypeChecker,PyUnresolvedReferences
def fixed_immediately(trajectory: pandas.Series, dlimit: float, flimit: float) -> bool:
	mask1 = (trajectory <= dlimit)
	mask2 = (trajectory >= flimit)
	return (mask1.sum() + mask2.sum()) == len(trajectory)


def fixed(trajectory: pandas.Series, flimit: float) -> bool:
	""" Tests whether the input series fixed at any point.
		If it throws a ValueError due to being an array, check the input data for duplicate labels
	"""
	try:
		return any(i > flimit for i in trajectory.values)
	except ValueError:
		message = "Encountered an error when selecting 'fixed' trajectories. This is usually caused by duplicate trajectory ids in the input table."
		if isinstance(trajectory, pandas.DataFrame):
			message += f"\nRename these trajectories: {list(trajectory.index)}"
		raise ValueError(message)
	except TypeError as exception:
		logger.warning(f"Trajectory: {trajectory.values}")
		logger.warning(f"flimit: {flimit}")

		raise exception


def only_fixed(trajectory: pandas.Series, dlimit: float, flimit: float) -> bool:
	""" Tests whether the series immediately fixed and stayed fixed."""
	series = ((i > flimit or i < dlimit) for i in trajectory.values)
	return all(series)


def overlap(left: pandas.Series, right: pandas.Series, dlimit: float) -> int:
	result = [(i > dlimit and j > dlimit) for i, j in zip(left.values, right.values)]

	return sum(result)


# noinspection PyTypeChecker
def find_boundaries_fixed(trajectory: pandas.Series, flimit: float) -> Optional[Tuple[int, int]]:
	fixed_series = trajectory[trajectory > flimit]
	# Assume index is sorted to avoid the situation where the index is str.
	if fixed_series.empty: return None
	else:
		return fixed_series.index[0], fixed_series.index[-1]


def _get_git_log() -> str:
	filename = Path(__file__).parent.parent / ".git" / "logs" / "HEAD"
	try:
		contents = filename.read_text()
	except FileNotFoundError:
		contents = ""
	return contents


def get_commit_hash() -> str:
	commit_hash = "n/a"
	contents = _get_git_log()
	if contents:
		contents = contents.split('\n')
		contents = [i.strip() for i in contents if i.strip()]
		reader = csv.reader(contents, delimiter = '\t')
		for line in reader:
			if line:
				hash_string = line[0]
				try:
					commit_hash = hash_string.split()[1]
				except IndexError:
					continue
		commit_hash = commit_hash[:7]
	else:
		commit_hash = "not available"
	return commit_hash


def get_pair_combinations(elements: Iterable[str]) -> Generator[Tuple[str, str], None, None]:
	# Use a generator to avoid using too much memory for large datasets.
	# The only reason to use a list was so tqdm could automatically figure out how many operations had to be done,
	# but this can be calculated manually by passing the total number of trajectories as an additional parameter.
	return itertools.combinations(elements, 2)


def calculate_number_of_combinations(n: int) -> int:
	""" Calculates the number of pairwise combinations for an iterable of size `elements` using the
		formula `n!/(n-r)!
		Since we only want combinations of pairs (`r` = 2), this can be simplified to
		`n * (n-1).
	"""
	return n * (n - 1)




from .generate_scripts import generate_r_script, execute_r_script
from .generate_tables import GGMuller, generate_trajectory_table
from .genotypecollection import Genotype, GenotypeCollection
from .import_file import *
from .import_tables import import_table
from .import_timeseries import parse_trajectory_table, parse_genotype_table
from .mullerformat import GenerateMullerDataFrame
from . import projectdata




import pandas


# noinspection PyTypeChecker,PyUnresolvedReferences
def _get_timepoint_above_threshold(transposed_timepoints: pandas.DataFrame, cutoff: float, name: str = None) -> pandas.Series:
	"""
		Calculates when a genotype was first fixed based on `cutoff`. The resulting series will be named `name`, if given.
	Parameters
	----------
	transposed_timepoints: pandas.DataFrame
		A dataframe indexed by timepoint and with each series a separate column.
	cutoff: float
		The threshold a frequency must exceed to be considered a valid timepoint.

	Returns
	-------
	pandas.Series
		A series mapping series name to the first timepoint that series first exceeded the threshold value.
		Series that never exceed the threshold are mapped to a value of 0.
	"""
	threshold_series: pandas.Series = (transposed_timepoints > cutoff).idxmax(0).sort_values()
	if name:
		threshold_series.name = name
	return threshold_series


def get_first_detected_timepoint(transposed_genotypes: pandas.DataFrame, cutoff: float) -> pandas.Series:
	"""

	Parameters
	----------
	transposed_genotypes:pandsa.DataFrame
		A table where rows correspond to timepoints and columns correspond to genotypes.
	cutoff

	Returns
	-------

	"""
	initial_genotype_values = transposed_genotypes.iloc[0].transpose()
	first_detected = _get_timepoint_above_threshold(transposed_genotypes, cutoff, 'firstDetected')
	first_detected_reduced = first_detected[(first_detected != 0) | (initial_genotype_values > cutoff)]

	return first_detected_reduced


def get_first_significant_timepoint(transposed_genotypes: pandas.DataFrame, cutoff: float) -> pandas.Series:
	"""
		Retrieves the first timepoint in each genotye that exceeds the significant cutoff.
	Parameters
	----------
	transposed_genotypes
	cutoff

	Returns
	-------
	"""

	initial_genotype_values = transposed_genotypes.iloc[0].transpose()
	first_above_threshold = _get_timepoint_above_threshold(transposed_genotypes, cutoff, 'firstSignificant')
	first_above_threshold_reduced_dict = dict()
	for key, value in first_above_threshold.items():
		initial_value = initial_genotype_values[key]
		if value == transposed_genotypes.index[0]:
			value = value if initial_value > cutoff else transposed_genotypes.index[-1]
		first_above_threshold_reduced_dict[key] = value

	first_above_threshold_reduced = pandas.Series(first_above_threshold_reduced_dict, name = 'firstThreshold')
	return first_above_threshold_reduced


def get_first_fixed_timepoint(transposed_genotypes: pandas.DataFrame, cutoff: float) -> pandas.DataFrame:
	"""
		Retrives the first timepoint every genotype exceeded the cutoff value.
	Parameters
	----------
	transposed_genotypes
	cutoff

	Returns
	-------

	"""
	first_fixed = _get_timepoint_above_threshold(transposed_genotypes, cutoff, 'firstFixed')
	if cutoff != 0:
		# If we are checking for the lowest frequency breakpoint, ignore this check.
		first_fixed_reduced: pandas.DataFrame = first_fixed.iloc[first_fixed.to_numpy().nonzero()]
	else:
		first_fixed_reduced = first_fixed.to_frame()

	return first_fixed_reduced



"""
Generates a palette using the information.
"""
from typing import Dict, List
import pandas
import random

try:
	from . import colorset
	from muller import treetools
except ModuleNotFoundError:
	from muller import treetools
	from muller.graphics.palettes import colorset


def apply_clade_colorscheme(clade: List[str], colorscheme: str) -> Dict[str, str]:
	"""
		Assigns colorschemes to each clade group
	Parameters
	----------
	clade: List[str]: A list of genotypes in this clade
	colorscheme: The name of a valid colorscheme

	Returns
	-------

	"""
	colorscheme = colorset.load_colorscheme(colorscheme, len(clade))
	palette = dict(zip(clade, colorscheme))
	return palette


def generate_lineage_palette(edges: pandas.Series) -> Dict[str, str]:
	clades = treetools.parse_tree(edges)
	major_clades = get_major_clades(clades)
	clades['majorClade'] = [major_clades[i] for i in clades.index]
	groups = clades.groupby(by = 'majorClade')

	palette = dict()
	_schemes = colorset.distinctive_colorschemes + colorset.distinctive_colorschemes + colorset.distinctive_colorschemes # In case there are a very large number of clades.
	for clade_label, colorscheme in zip(clades['majorClade'].unique(), _schemes):
		if colorscheme is None:
			colorscheme = random.choice(colorset.distinctive_colorschemes)
		group = groups.get_group(clade_label)
		clade_palette = apply_clade_colorscheme(group.index, colorscheme)
		palette.update(clade_palette)
	return palette


def get_major_clades(tree_table: pandas.DataFrame):
	clade_counts = tree_table['Parent'].value_counts()
	cutoff = int(len(tree_table) / 20)
	major_clades = list(clade_counts[clade_counts > cutoff].index)
	major_clade_map = dict()
	for child in tree_table.index:
		if child in major_clades:
			first_major_clade = child
		else:
			parents = treetools.get_parent_nodes(tree_table, child)
			try:
				first_major_clade = [i for i in parents if i in major_clades][0]
			except IndexError:
				first_major_clade = child
		major_clade_map[child] = first_major_clade
	return major_clade_map



from pathlib import Path
from typing import *

import pandas


def read_map(contents: Union[None, str, Path]) -> Dict[str, str]:
	"""
		Parses the contents of any file used to specify a key-value pair. This is used when defining palettes and gene aliases.
	Parameters
	----------
	contents: Union[str,Path]
		Either a path to the file or the contents of the file.

	Returns
	-------
	Dict[str,str]
	"""
	if contents is None: return dict()
	lines = readfile(contents)

	filemap = dict()
	for line in lines:
		fields = line.split('\t')
		try:
			# We only care about the first two values in the line. Anything after that is ignored. May be used to add notes ot the file.
			# It's fine if the first row contains headers. They shouldn't interfere with any of the genotype/genes/etc.
			left, right, *_ = fields
		except ValueError:
			# Not enough values in line.
			continue
		left = left.strip()
		right = right.strip()
		if left and right:
			filemap[left] = right

	return filemap


def readfile(contents: Union[str, Path]) -> List[str]:
	"""
		Takes a file path or the contents of a file and attemts to split it into lines.
	Parameters
	----------
	contents: Union[str, Path]

	Returns
	-------
	List[str]
		The lines in the file.

	"""
	if not isinstance(contents, str):
		contents = contents.read_text()
	lines = contents.split('\n')  # Use this instead of splitlines sine splitlines sometimes doesn't work right.
	# Remove any whitespace to the left and right of the contents in the line. This should avaid issues with extra whitespace padding on the lines.
	clean_lines = [l.strip() for l in lines]
	# Remove empty lines
	nonempty_lines = [l for l in clean_lines if l]
	return nonempty_lines


def parse_known_genotypes(known_genotypes: Union[str, Path]) -> List[List[str]]:
	lines = readfile(known_genotypes)

	genotypes = list()
	for line in lines:
		trajectories = line.split(',')
		trajectories = [i.strip() for i in trajectories]
		genotypes.append(trajectories)
	return genotypes


def _get_column_values(table: pandas.DataFrame, column: str) -> List[str]:
	if column not in table.columns:
		if column.title() in table.columns:
			column = column.title()
		else:
			return []

	column_values = table[column].tolist()
	column_values = [(i if isinstance(i, str) else "") for i in column_values]
	return column_values





from pathlib import Path
import pandas
from typing import Dict, List, Union, Optional, Mapping
import re

from muller.dataio import read_map

def read_genotype_annotations(info: pandas.DataFrame, genotype_members: Dict[str, List[str]], alias_filename: Optional[Path] = None) -> Dict[
	str, List[str]]:
	if info is not None:
		genotype_annotations = parse_genotype_annotations(
			genotype_members,
			info,
			alias_filename
		)
	else:
		genotype_annotations = {}
	return genotype_annotations



def _clean_gene_label(label: str) -> str:
	if not isinstance(label, str):
		return ""
	gene_separator = "/>"
	if gene_separator in label:
		sublabels = [_clean_gene_label(i) for i in label.split(gene_separator)]
		clean_label = "|".join(sublabels)
	else:
		pattern = "[a-zA-Z_0-9\s]+"
		match = re.match(pattern, label)
		if match:
			clean_label = label
			#clean_label = match.group(0)
		else:
			clean_label = ""
	if clean_label.endswith('<'):
		clean_label = clean_label[:-1]
	return clean_label


def _clean_annotation_label(label: str) -> str:
	"""
		Removes extra stuff included in the annotation form breseq.
	"""
	if not isinstance(label, str): return ""
	if 'intergenic' in label:
		result = 'intergenic'
	elif 'del' in label:
		result = 'del'
	else:
		# Based on a specific format.
		result = label.split('(')[0]
		if result == "":
			result = label
	return result


def _extract_value(row: Mapping[str, str], column: str) -> Optional[str]:
	"""
		Extracts a value from a Mapping (usually a pandas.Series object) using a few variations of the input column.
		For example, the column containing gene labels could be names 'Gene' or 'gene'. pandas.Series objects save null values as math.nan, a float.
	"""
	if column in row:
		value = row[column]
	elif column.lower() in row:
		value = row[column.lower()]
	elif column.title() in row:
		value = row[column.title()]
	else:
		value = None
	if isinstance(value, float):
		value = None
	return value


def extract_annotations(info: pandas.DataFrame, alias_filename: Optional[Path] = None) -> Dict[str, str]:
	""" Maps trajectories to their corresponding mutation.
		Parameters
		----------
		info: pandas.DataFrame
			The info table. Each column label should be lowercase.
		alias_filename: Will map each gene from the infotable to an alias.
	"""
	# Make the column labels lowercase so we don't have to care about capitalization
	info.columns = [str(i).lower() for i in info.columns]
	if alias_filename:
		alias_map = read_map(alias_filename)
	else:
		alias_map = {}

	column_label_gene = 'gene'
	annotation_column = 'annotation'

	trajectory_annotations: Dict[str,str] = dict()
	for trajectory_label, row in info.iterrows():
		# Since the columns from the info table should already be lowercase, we don't need this method anymore.
		gene_value = row.get(column_label_gene)
		# Need to clean up the gene label.
		gene = _clean_gene_label(gene_value)
		# Rename the gene according to the alias map.
		gene = alias_map.get(gene, gene)

		annotation_value = _extract_value(row, annotation_column)
		annotation = _clean_annotation_label(annotation_value)

		result = list()
		if gene:
			result.append(gene)
		if annotation:
			result.append(annotation)
		trajectory_annotations[trajectory_label] = " ".join(result)
	return trajectory_annotations


def parse_genotype_annotations(genotype_members: Mapping[str, Union[str, List[str]]], info: pandas.DataFrame,
		alias_filename: Optional[Path] = None) -> Dict[
	str, List[str]]:
	"""
		Parses the non-numeric tables form the input table and attempts to map annotations to genotypes.
	Parameters
	----------
	genotype_members: Mapping[str, List[str]]
	info: pandas.DataFrame
		The columns from the input table that do not contain frequency measurements. The annotations will be extracted from these columns, if available:
		- `gene`
		- `annotation`

	alias_filename
	"""
	trajectory_annotations: Dict[str,str] = extract_annotations(info, alias_filename)

	genotype_annotations = dict()
	for genotype_label, members in genotype_members.items():
		# Make sure the members are not stored as a string.
		if isinstance(members, str):
			members = members.split('|')
		# Get all annotations relevant to this genotype.
		# Want to combine the annotation for each trajectory into a single str.
		member_values = list()
		for i in members:
			trajectory_annotation = trajectory_annotations.get(i, "")
			if trajectory_annotation:
				if isinstance(trajectory_annotation, str):
					member_values.append(trajectory_annotation)
				else:
					member_values.append(" ".join(trajectory_annotation))
		# Remove missing annotations
		# Also remove excess whitespace
		member_values: List[str] = [i.strip() for i in member_values if i]
		genotype_annotations[genotype_label] = member_values
	return genotype_annotations

if __name__ == "__main__":
	pass


from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
# plt.switch_backend('agg')
import pandas
import seaborn

def plot_heatmap(data: pandas.DataFrame, filename: Path):
	font = {
		'size': 20
	}
	use_annotations = len(data) < 20  # So the annotations are actually visible
	matplotlib.rc('font', **font)
	figsize = (20, 20)
	fig, ax = plt.subplots(figsize = figsize)
	seaborn.heatmap(
		data,
		ax = ax,
		annot = use_annotations,
		cmap = 'Reds',
		square = True,
		xticklabels = True,
		yticklabels = True
	)
	ax.tick_params(axis = 'both', which = 'major', labelsize = 24)
	ax.set_ylabel("Trajectory Label", size = 32)
	ax.set_xlabel("Trajectory Label", size = 32)
	ax.set_title("p-values of all mutational trajectories", size = 36)
	plt.tight_layout()
	fig.savefig(str(filename), format = 'svg')



from pathlib import Path
from typing import List, Optional

import matplotlib.pyplot as plt
import seaborn

# plt.style.use("/home/cld100/Documents/sandbox/matplotlibrc")


def generate_distance_plot(distances: List[float], similarity_cutoff: float, filename: Optional[Path] = None, ax: plt.Axes = None):
	""" Shows the spread of computed pairwise distances as a histogram."""
	if not ax:
		fig, ax = plt.subplots(figsize = (12, 10))

	seaborn.distplot(distances, ax = ax, kde = False, rug = True, bins = 20)
	ax.axvline(similarity_cutoff, color = 'red')

	ax.set_title("Pairwise distances between each pair of trajectories")
	ax.set_xlabel("Distance")
	ax.set_ylabel("Count")
	ax.set_xlim(0, max(distances))
	plt.tight_layout()
	if filename:
		plt.savefig(filename)
	else:
		plt.show()



from setuptools import setup
from muller import commandline_parser
from pathlib import Path
if commandline_parser.DEBUG:
	message = "The scripts are currently in debug mode!"
	raise ValueError(message)
FOLDER = Path(__file__).parent
README = FOLDER / "README.md"

with README.open() as readmefile:
	LONG_DESCRIPTION = readmefile.read()

setup(
	name = 'lolipop',
	version = commandline_parser.__VERSION__,
	packages = [
		'muller', 'muller.clustering', 'muller.dataio', 'muller.graphics', 'muller.inheritance',
		'muller.graphics.palettes', 'muller.clustering.metrics'
	],
	extras_require = {
		'Show progressbar for large datasets': ['tqdm'],
		'Additional support for parsing files': ['beautifulsoup4'],
		'List similar annotations when selecting genotypes': ['fuzzywuzzy'],
		'Generate composite graphs from the genotye/lineage/timeseries plots': ["pillow"],
		'To run tests': ["pytest"]
	},
	provides = 'lolipop',
	url = 'https://github.com/cdeitrick/lolipop',
	license = 'MIT',
	author = 'chris deitrick',
	author_email = 'chrisdeitrick1@gmail.com',
	description = 'A set of scripts to cluster mutational trajectories into genotypes and infer lineage in the context of population evolution experiments.',
	long_description = LONG_DESCRIPTION,
	long_description_content_type='text/markdown',
	install_requires = [
		'pandas>=0.24.0', 'loguru', 'scipy>=1.3.0', 'matplotlib>=3.0.0','graphviz',
		'pygraphviz', 'seaborn', 'numpy>=1.16.2', 'xlrd', 'shapely>=1.6.4'
	],
	tests_requires = ['pytest'],
	classifiers = [
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	scripts = ["lolipop"]
)
